import json
import logging
import httpx

from ..exceptions import ClientException, ApiException
from ..const import API_URL

LOG = logging.getLogger(__package__)

class NgenicBase(object):
    """Superclass for all models"""

    def __init__(self, session=None, json=None):
        """Initialize our base object.

        :param session:
            (required) httpx client
        :param json:
            (required) Json representation of the concrete model
        """

        # store the httpx session on each object so we can reuse the connection
        self._session = session

        # backing json of the model
        self._json = json
    
    def json(self):
        """Get a json representaiton of the model

        :return:

        """
        return self._json

    def uuid(self):
        """Get uuid attribute"""
        return self["uuid"]

    def __setitem__(self, attribute, data):
        self._json[attribute] = data

    def __getitem__(self, attribute):
        if attribute not in self._json:
            raise AttributeError(attribute)
        return self._json[attribute]

    def update(self):
        raise ClientException("Cannot update a '%s'" % self.__class__.__name__)

    def _parse(self, response):
        rsp_json = None

        if response is None:
            return None

        if response.status_code == 204:
            return None

        try:
            rsp_json = response.json()
        except ValueError:
            raise ApiException("Ngenic API return an invalid json body (status=%d)" % response.status_code)

        return rsp_json

    def _new_instance(self, instance_class, json, **kwargs):
        """Create a new model instance

        :param class instance_class:
            (required) class of instance to initialize with json
        :param dict json:
            (required) data to initialize the instance with
        :param kwargs:
            Additional data required by the instance type
        :return:
            new `instance_class` or `list(instance_class)`
        """
        if json is not None and (not isinstance(json, dict) and not isinstance(json, list)):
            raise ClientException("Invalid data to create new instance with (expected json)")
        if not json:
            return None

        if isinstance(json, list):
            return list(instance_class(session=self._session, json=inst_json, **kwargs) for inst_json in json)
        else:
            return instance_class(session=self._session, json=json, **kwargs)

    def _parse_new_instance(self, url, instance_class, **kwargs):
        """Get JSON from an URL and create a new instance of it

        :param str url:
            (required) url to get instance data from
        :param class instance_class:
            (required) class of instance to initialize with parsed data
        :param kwargs:
            may contain additional args to the instance class
        :return:
            new `instance_class`
        :rtype:
            `instance_class`
        """
        ret_json = self._parse(self._get(url))
        return self._new_instance(instance_class, ret_json, **kwargs)

    async def _async_parse_new_instance(self, url, instance_class, **kwargs):
        """Get JSON from an URL and create a new instance of it

        :param str url:
            (required) url to get instance data from
        :param class instance_class:
            (required) class of instance to initialize with parsed data
        :param kwargs:
            may contain additional args to the instance class
        :return:
            new `instance_class`
        :rtype:
            `instance_class`
        """
        ret_json = self._parse(await self._async_get(url))
        return self._new_instance(instance_class, ret_json, **kwargs)

    def _request(self, method, *args, **kwargs):
        """Make a HTTP request.
        This is the generic method for all requests, it will handle errors etc in a common way.

        :param str method:
            (required) HTTP method (i.e get, post, delete)
        :param args:
            Additional args to requests lib
        :param kwargs:
            Additional kwargs to requests lib
        :return:
            request
        """
        r = None
        try:
            if not isinstance(self._session, httpx.Client):
                raise ValueError("Cannot use sync methods when context is async")

            request_method = getattr(self._session, method)
            r = request_method(*args, **kwargs)
            
            # raise for e.g. 401
            r.raise_for_status()

            return r
        except httpx.HTTPError as exc:
            raise ClientException(self._get_error("A request exception occurred", r, parent_ex=exc))
        except Exception as exc:
            raise ClientException(self._get_error("An exception occurred", r, parent_ex=exc))

    async def _async_request(self, method, *args, **kwargs):
        """Make a HTTP request (async).
        This is the generic method for all requests, it will handle errors etc in a common way.

        :param str method:
            (required) HTTP method (i.e get, post, delete)
        :param args:
            Additional args to requests lib
        :param kwargs:
            Additional kwargs to requests lib
        :return:
            request
        """
        r = None
        try:
            if not isinstance(self._session, httpx.AsyncClient):
                raise ValueError("Cannot use async methods when context is sync")

            request_method = getattr(self._session, method)
            r = await request_method(*args, **kwargs)
            
            # raise for e.g. 401
            r.raise_for_status()

            return r
        except httpx.HTTPError as exc:
            raise ClientException(self._get_error("A request exception occurred", r, parent_ex=exc))
        except Exception as exc:
            raise ClientException(self._get_error("An exception occurred", r, parent_ex=exc))

    def _get_error(self, msg, req, parent_ex=None):
        if req is not None and req.status_code == 429:
            # Too many requests
            server_msg = "Too many requests have been made, retry again after %s" % req.headers["X-RateLimit-Reset"]
        else:
            try:
                server_msg = req.json()["message"]
            except:
                if req is not None:
                    server_msg = str(req.status_code)
                elif parent_ex is not None:
                    if isinstance(parent_ex, httpx.ConnectTimeout):
                        server_msg = "Timed out connecting to ngenic server"
                        
                    elif isinstance(parent_ex, httpx.ConnectTimeout):
                        server_msg = "Timed out sending request to ngenic server"
                    else:
                        server_msg = str(parent_ex)
                else:
                    server_msg = "Unknown error"
                pass

        return "%s: %s" % (msg, server_msg)

    def _prehandle_write(self, data, is_json, **kwargs):
        headers = {}

        if is_json:
            data = json.dumps(data) if data is not None else data
            headers["Content-Type"] = "application/json"

        if "headers" in kwargs:
            # let caller override headers
            headers.update(kwargs.get("headers"))

        return (data, headers)

    def _delete(self, url, **kwargs):
        LOG.debug("DELETE %s with %s", url, kwargs)
        return self._request("delete",
                             "%s/%s" % (API_URL, url))
                            
    def _async_delete(self, url, **kwargs):
        LOG.debug("DELETE %s with %s", url, kwargs)
        return self._async_request("delete",
                             "%s/%s" % (API_URL, url))

    def _get(self, url, **kwargs):
        LOG.debug("GET %s with %s", url, kwargs)
        return self._request("get",
                             "%s/%s" % (API_URL, url),
                             **kwargs)

    async def _async_get(self, url, **kwargs):
        LOG.debug("GET %s with %s", url, kwargs)
        return await self._async_request("get",
                             "%s/%s" % (API_URL, url),
                             **kwargs)

    def _post(self, url, data=None, is_json=True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, kwargs)

        LOG.debug("POST %s with %s, %s", url, data, kwargs)
        return self._request("post",
                             "%s/%s" % (API_URL, url),
                             data=data,
                             headers=headers)

    def _async_post(self, url, data=None, is_json=True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, kwargs)

        LOG.debug("POST %s with %s, %s", url, data, kwargs)
        return self._async_request("post",
                             "%s/%s" % (API_URL, url),
                             data=data,
                             headers=headers)

    def _put(self, url, data=None, is_json=True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        
        LOG.debug("PUT %s with %s, %s", url, data, kwargs)
        return self._request("put",
                             "%s/%s" % (API_URL, url),
                             data=data,
                             headers=headers)

    def _async_put(self, url, data=None, is_json=True, **kwargs):
        data, headers = self._prehandle_write(data, is_json, **kwargs)
        
        LOG.debug("PUT %s with %s, %s", url, data, kwargs)
        return self._async_request("put",
                             "%s/%s" % (API_URL, url),
                             data=data,
                             headers=headers)



import json
from .base import NgenicBase
from .room import Room
from .node import Node
from ..const import API_PATH

class Tune(NgenicBase):
    def __init__(self, session, json):
        super(Tune, self).__init__(session=session, json=json)

    def uuid(self):
        """Get the tune UUID"""

        # If a tune was fetched with the list API, it contains "tuneUuid"
        # If it was fetched directly (with UUID), it contains "uuid"
        try:
            return self["tuneUuid"]
        except AttributeError:
            return super().uuid()

    def rooms(self):
        """List all Rooms associated with a Tune. A Room contains an indoor sensor.

        :return:
            a list of rooms
        :rtype:
            `list(~ngenic.models.room.Room)`
        """
        url = API_PATH["rooms"].format(tuneUuid=self.uuid(), roomUuid="")
        return self._parse_new_instance(url, Room, tune=self)

    async def async_rooms(self):
        """List all Rooms associated with a Tune (async). A Room contains an indoor sensor.

        :return:
            a list of rooms
        :rtype:
            `list(~ngenic.models.room.Room)`
        """
        url = API_PATH["rooms"].format(tuneUuid=self.uuid(), roomUuid="")
        return await self._async_parse_new_instance(url, Room, tune=self)

    def room(self, roomUuid):
        """Get data about a Room. A Room contains an indoor sensor.

        :param str roomUuid:
            (required) room UUID
        :return:
            the room
        :rtype:
            `~ngenic.models.room.Room`
        """
        url = API_PATH["rooms"].format(tuneUuid=self.uuid(), roomUuid=roomUuid)
        return self._parse_new_instance(url, Room, tune=self)

    async def async_room(self, roomUuid):
        """Get data about a Room (async). A Room contains an indoor sensor.

        :param str roomUuid:
            (required) room UUID
        :return:
            the room
        :rtype:
            `~ngenic.models.room.Room`
        """
        url = API_PATH["rooms"].format(tuneUuid=self.uuid(), roomUuid=roomUuid)
        return await self._async_parse_new_instance(url, Room, tune=self)

    async def async_room(self, roomUuid):
        """Get data about a Room (async). A Room contains an indoor sensor.

        :param str roomUuid:
            (required) room UUID
        :return:
            the room
        :rtype:
            `~ngenic.models.room.Room`
        """
        url = API_PATH["rooms"].format(tuneUuid=self.uuid(), roomUuid=roomUuid)
        return await self._async_parse_new_instance(url, Room, tune=self)

    def nodes(self):
        """List all Nodes associated with a Tune. A Node is a logical network entity.

        :return:
            a list of nodes
        :rtype:
            `list(~ngenic.models.node.Node)`
        """
        url = API_PATH["nodes"].format(tuneUuid=self.uuid(), nodeUuid="")
        return self._parse_new_instance(url, Node, tune=self)

    async def async_nodes(self):
        """List all Nodes associated with a Tune (async). A Node is a logical network entity.

        :return:
            a list of nodes
        :rtype:
            `list(~ngenic.models.node.Node)`
        """
        url = API_PATH["nodes"].format(tuneUuid=self.uuid(), nodeUuid="")
        return await self._async_parse_new_instance(url, Node, tune=self)

    def node(self, nodeUuid):
        """Get data about a Node. A Node is a logical network entity.

        :param str nodeUuid:
            (required) node UUID
        :return:
            the node
        :rtype:
            `~ngenic.models.node.Node`
        """
        url = API_PATH["nodes"].format(tuneUuid=self.uuid(), nodeUuid=nodeUuid)
        return self._parse_new_instance(url, Node, tune=self)

    async def async_node(self, nodeUuid):
        """Get data about a Node (async). A Node is a logical network entity.

        :param str nodeUuid:
            (required) node UUID
        :return:
            the node
        :rtype:
            `~ngenic.models.node.Node`
        """
        url = API_PATH["nodes"].format(tuneUuid=self.uuid(), nodeUuid=nodeUuid)
        return await self._async_parse_new_instance(url, Node, tune=self)



"""This module contains the main interfaces to the API."""

from .models import NgenicBase
from .models import Tune
from .models import Room
from .const import API_PATH
from .const import API_URL
from .exceptions import ApiException, ClientException

import logging
import json
import httpx

LOG = logging.getLogger(__package__)

# 30sec for connect, 10sec elsewhere.
timeout = httpx.Timeout(10.0, connect=2.0)

class BaseClient(NgenicBase):
    def __init(self, session):
        super(BaseClient, self).__init__(session=session)

    def tunes(self):
        """Fetch all tunes

        :return:
            a list of tunes
        :rtype:
            `list(~ngenic.models.tune.Tune)`
        """
        url = API_PATH["tunes"].format(tuneUuid="")
        return self._parse_new_instance(url, Tune)

    async def async_tunes(self):
        """Fetch all tunes (async)

        :return:
            a list of tunes
        :rtype:
            `list(~ngenic.models.tune.Tune)`
        """
        url = API_PATH["tunes"].format(tuneUuid="")
        return await self._async_parse_new_instance(url, Tune)

    def tune(self, tuneUuid):
        """Fetch a single tune

        :param str tuneUUid:
            (required) tune UUID
        :return:
            the tune
        :rtype:
            `~ngenic.models.tune.Tune`
        """
        url = API_PATH["tunes"].format(tuneUuid=tuneUuid)
        return self._parse_new_instance(url, Tune)

    def async_tune(self, tuneUuid):
        """Fetch a single tune

        :param str tuneUUid:
            (required) tune UUID
        :return:
            the tune
        :rtype:
            `~ngenic.models.tune.Tune`
        """
        url = API_PATH["tunes"].format(tuneUuid=tuneUuid)
        return self._async_parse_new_instance(url, Tune)
    
class Ngenic(BaseClient):
    def __init__(self, token):
        """Initialize an ngenic object.

        :param token:
            (required) OAuth2 bearer token
        """

        # this will be added to the HTTP Authorization header for each request
        self._token = token

        # this header will be added to each HTTP request
        self._auth_headers = {"Authorization": "Bearer %s" % self._token}

        session = httpx.Client(headers=self._auth_headers, timeout=timeout) 

        # initializing this doesn't require a session or json
        super(Ngenic, self).__init__(session=session)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """Close the session if it was not created as a context manager"""
        self._session.close()

class AsyncNgenic(BaseClient):
    def __init__(self, token):
        """Initialize an async ngenic object.

        :param token:
            (required) OAuth2 bearer token
        """

        # this will be added to the HTTP Authorization header for each request
        self._token = token

        # this header will be added to each HTTP request
        self._auth_headers = {"Authorization": "Bearer %s" % self._token}

        session = httpx.AsyncClient(headers=self._auth_headers, timeout=timeout) 

        # initializing this doesn't require a session or json
        super(AsyncNgenic, self).__init__(session=session)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.async_close()

    async def async_close(self):
        """Close the session if it was not created as a context manager"""
        await self._session.aclose()



import json
from enum import Enum
from .base import NgenicBase
from ..const import API_PATH

# Undocumented in API
class MeasurementType(Enum):
    TEMPERATURE = "temperature_C"
    TARGET_TEMPERATURE = "target_temperature_C"
    HUMIDITY = "humidity_relative_percent"
    CONTROL_VALUE = "control_value_C"
    POWER_KW = "power_kW"
    ENERGY_KWH = "energi_kWH"

class Measurement(NgenicBase):
    def __init__(self, session, json, node, measurement_type):
        self._parentNode = node
        self._measurementType = measurement_type

        super(Measurement, self).__init__(session=session, json=json)

    def get_type(self):
        """Get the measurement type

        :return:
            measurement type
        :rtype:
            `~ngenic.models.measurement.MeasurementType`
        """
        return self._measurementType



API_URL = "https://app.ngenic.se/api/v3"

API_PATH = {
  "tunes":                  "tunes/{tuneUuid}",
  "rooms":                  "tunes/{tuneUuid}/rooms/{roomUuid}",
  "nodes":                  "tunes/{tuneUuid}/gateway/nodes/{nodeUuid}",
  "node_status":            "tunes/{tuneUuid}/nodestatus",
  "measurements":           "tunes/{tuneUuid}/measurements/{nodeUuid}",
  "measurements_types":     "tunes/{tuneUuid}/measurements/{nodeUuid}/types",
  "measurements_latest":    "tunes/{tuneUuid}/measurements/{nodeUuid}/latest"
}



import json
from .base import NgenicBase
from ..const import API_PATH

class Room(NgenicBase):
    def __init__(self, session, json, tune):
        self._parentTune = tune

        super(Room, self).__init__(session=session, json=json)

    def update(self):
        """Update this room with its current values"""
        roomUuid = self["uuid"]

        url = API_PATH["rooms"].format(tuneUuid=self._parentTune.uuid(), roomUuid=roomUuid)
        self._put(url, data=self.json())

    async def async_update(self):
        """Update this room with its current values (async)"""
        roomUuid = self["uuid"]

        url = API_PATH["rooms"].format(tuneUuid=self._parentTune.uuid(), roomUuid=roomUuid)
        await self._async_put(url, data=self.json())
